#!/usr/bin/env python3
import aputils
import argparse
import json
import socket

from blib import JsonBase
from functools import cached_property, lru_cache
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer
from pathlib import Path
from urllib.error import HTTPError
from urllib.request import Request, urlopen


HOME = f"""<!DOCTYPE html>
<html>
	<head>
		<title>ActivityPub Utilities Signature Verifier</title>
		<meta name="viewport" content="width=device-width, initial-scale=1">
		<style>
			a {{
				color: #A8D;
			}}

			a:hover {{
				color: #A6D;
			}}

			body {{
				color: white;
				background: #222;
				font-family: sans-serif;
				margin: 0px;
			}}

			h1 {{
				text-align: center;
				color: #A6D;
			}}

			p:first-child {{
				margin-top: 0px;
			}}

			p:last-child {{
				margin-botton: 0px;
			}}

			pre {{
				border: 1px solid transparent;
				border-radius: 5px;
				background: #403344;
				padding: 5px;
				overflow-x: auto;
			}}

			.section {{
				border: 1px solid transparent;
				border-radius: 5px;
				padding: 5px;
				background: #333;
				margin: 10px;
			}}
		</style>
	</head>
	<body>
		<h1 class="section">ActivityPub Utilities Signature Verifier</h1>

		<div class="section">
			<p>
				This server allows you to test your HTTP signature implementation. Just send a request
				of any method to an endpoint that is not any of the following:
			</p>
			<ul>
				<li>GET /</li>
				<li>GET /actor</li>
				<li>GET /.well-known/webfinger</li>
			</ul>
		</div>

		<div class="section">
			<p>
				The server will respond with a JSON message that includes some info sent to the server
				as well as whether or not the signature is valid. Example:
			</p>
			<pre>
{{
	"status": 200,
	"message": "HTTP signature is valid :3",
	"method": "GET",
	"path": "/heck",
	"address": "192.168.2.5",
	"valid": true,
	"headers": {{
		"Host": "valtest.barkshark.xyz",
		"User-Agent": "http.rb/5.1.1 (Mastodon/4.3.0-alpha.3+glitch; +https://barkshark.xyz/; im gay)",
		"Accept": "application/activity+json",
		"Accept-Encoding": "identity",
		"Content-Type": "application/activity+json",
		"Date": "Wed, 03 Apr 2024 17:24:50 GMT",
		"Signature": "[didn't wanna include this because it's so damn long]"
	}}
}}</pre>
		</div>

		<p class="section">
			Powered by
			<a href="https://git.barkshark.xyz/barkshark/aputils">ActivityPub Utilities/{aputils.__version__}</a>
		</p>
	</body>
</html>
""" # noqa: E501

ACTOR_BIO = """<p>Signature validation actor</p>
<p>Have an HTTP signature implementation you want to test? Just send a request to me. See more info
at <a href="{url}/">{url}</a>"""


def get_actor() -> JsonBase:
	return JsonBase({
		"@context": [
			"https://www.w3.org/ns/activitystreams"
		],
		"type": "Application",
		"id": f"{URL}/actor",
		"preferredUsername": "valtest",
		"name": "Signature Verifier",
		"summary": ACTOR_BIO.format(url = URL),
		"manuallyApprovesFollowers": True,
		"inbox": f"{URL}/inbox",
		"url": f"{URL}/",
		"endpoints": {
			"sharedInbox": f"{URL}/inbox"
		},
		"publicKey": {
			"id": keyid,
			"owner": f"{URL}/actor",
			"publicKeyPem": signer.pubkey
		}
	})


def get_machine_addr() -> str:
	with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
		try:
			sock.connect(("10.254.254.255", 1))
			return sock.getsockname()[0]

		except Exception:
			return "127.0.0.1"


class ClientError(Exception):
	...


class RequestHandler(BaseHTTPRequestHandler):
	default_request_version = "HTTP/1.1"
	signature: aputils.Signature
	actor: aputils.Message
	signer: aputils.Signer


	@property
	def content_length(self) -> int:
		try:
			return int(self.headers.get("Content-Length", 0))

		except ValueError:
			raise ClientError(f"Actor is larger than {args.size_limit} bytes") from None


	@property
	def method(self) -> str:
		return self.command.upper()


	@property
	def remote(self) -> str:
		return self.headers.get("X-Real-Ip", self.headers.get("X-Forwarded-For", self.address_string()))


	@cached_property
	def parsed_headers(self) -> dict[str, str]:
		headers = {}

		for key, value in self.headers.items():
			key = key.title()

			if key.startswith(("X-Forwarded", "X-Real")):
				continue

			headers[key] = value

		return headers


	def body(self) -> bytes:
		return self.rfile.read(self.content_length or -1)


	def log_request(self, status: int, length: int = 0) -> None: # type: ignore
		path = self.path.split("?", 1)[0]
		agent = self.headers.get("User-Agent", "n/a")
		message = f"{self.remote} \"{self.method} {path}\" {status} {length} {agent}"

		if not args.skip_date_log:
			date = self.date_time_string()
			message = f"[{date}] " + message

		print(message, flush = True)


	def send(self,
			status: int,
			message: JsonBase | str,
			headers: dict[str, str] | None = None) -> None:

		if isinstance(message, str):
			data = message.encode("utf-8")

		else:
			data = message.to_json().encode("utf-8")

		if headers is None:
			headers = {"Content-Type": "application/json"}

		if "Content-Type" not in headers:
			headers["Content-Type"] = "application/json"

		self.log_request(status, len(data))

		self.send_response_only(status, None)
		self.send_header("Server", self.version_string())
		self.send_header("Date", self.date_time_string())

		for key, value in headers.items():
			self.send_header(key, value)

		self.end_headers()

		self.wfile.write(data)
		self.wfile.flush()


	def send_error(self, status: int, message: str) -> None: # type: ignore
		response = JsonBase({
			"status": status,
			"message": message,
			"method": self.method,
			"path": self.path.split("?", 1)[0],
			"query": self.path.split("?", 1)[1],
			"address": self.remote,
			"valid": False,
			"headers": self.parsed_headers
		})

		self.send(status, response, {"Content-Type": "application/json"})


	def handle_webfinger(self) -> None:
		try:
			path, query = self.path.split("?")

		except ValueError:
			self.send_error(400, "Missing query (ex. resource=acct:valtest@valtest.barkshark.xyz)")
			return

		if query != f"resource=acct:valtest@{args.hostname}":
			self.send_error(404, "Invalid user")
			return

		self.send(200, JsonBase({
			"subject": f"acct:valtest@{args.hostname}",
			"aliases": [
				f"{URL}/actor"
			],
			"links": [
				{
					"rel": "self",
					"type": "application/activity+json",
					"href": f"{URL}/actor"
				}
			]
		}))


	@lru_cache(maxsize = 1024, typed = True)
	def fetch_actor(self, url: str) -> aputils.Message:
		try:
			url, _ = url.split("#", 1)

		except ValueError:
			pass

		request = signer.sign_request(Request(
			url,
			method = "GET", headers = {
				"User-Agent": f"ApUtils Signature Verifier ({URL})"
			}
		))

		try:
			with urlopen(request) as response:
				if (length := int(response.headers.get("Content-Length", 0))) <= 0:
					raise ClientError("Actor body length is 0")

				if length > args.size_limit:
					raise ClientError(f"Actor is larger than {args.size_limit} bytes")

				return aputils.Message.parse(response.read())

		except HTTPError as error:
			msg = f"Failed to fetch actor: Status={error.code} Message='{str(error.read())}'"
			raise ClientError(msg) from None

		except json.JSONDecodeError as error:
			raise ClientError(f"Failed to parse actor: {str(error)}") from None

		except ValueError:
			raise ClientError("Content-Length header is not an integer") from None


	def parse_request(self) -> bool:
		if not BaseHTTPRequestHandler.parse_request(self):
			return False

		path = self.path.split("?", 1)[0]

		if self.method == "GET":
			if path == "/":
				self.send(200, HOME, {"Content-Type": "text/html"})
				return False

			if path == "/actor":
				self.send(200, get_actor(), {"Content-Type": "application/activity+json"})
				return False

			if path == "/.well-known/webfinger":
				self.handle_webfinger()
				return False

		if self.content_length > args.size_limit:
			self.send_error(400, f"Incoming message is larger than {args.size_limit} bytes")
			return False

		if self.method == "GET" and self.content_length:
			self.send_error(400, "GET messages should not have a body")
			return False

		if self.method in {"POST", "PUT"} and self.content_length <= 0:
			self.send_error(400, f"{self.method} messages should have a body")
			return False

		try:
			self.signature = aputils.Signature.new_from_headers(self.parsed_headers)

		except KeyError:
			self.send_error(400, "Missing signature header")
			return False

		try:
			self.actor = self.fetch_actor(self.signature.keyid)

		except ClientError as error:
			self.send_error(400, str(error))
			return False

		self.signer = aputils.Signer.new_from_actor(self.actor)

		if self.command.upper() == "GET" and self.content_length:
			self.send_error(400, "'GET' messages should not have a body")
			return False

		try:
			self.signer.validate_signature(self.method, path, self.parsed_headers)

		except aputils.SignatureFailureError as error:
			self.send_error(401, str(error))
			return False

		response = JsonBase({
			"status": 200,
			"message": "HTTP signature is valid :3",
			"method": self.method,
			"path": self.path.split("?", 1)[0],
			"query": self.path.split("?", 1)[1],
			"address": self.remote,
			"valid": True,
			"headers": self.parsed_headers
		})

		self.send(200, response)

		return False


parser = argparse.ArgumentParser(
	prog = "aputils",
	description = "Starts a server for validating HTTP signatures"
)

parser.add_argument("--hostname", "-n",
	help = "Domain or address this server will be hosted on (defaults to --addr value)")

parser.add_argument("--addr", "-a", default = "0.0.0.0",
	help = "IP address to listen on")

parser.add_argument("--port", "-p", default = 8080, type = int,
	help = "TCP port to listen on")

parser.add_argument("--size-limit", "-s", default = 1024 * 1024, type = int,
	help = "Max size of incoming or outgoing messages in bytes")

parser.add_argument("--protocol", "-r", default = "http", choices = ["https", "http"],
	help = "Supported protocol to advertise via the actor at /actor")

parser.add_argument("--skip-date-log", "-d", action = "store_true", default = False,
	help = "Don't append the date to the start of the access log lines")

args = parser.parse_args()

if args.hostname is None:
	args.hostname = args.addr if args.addr != "0.0.0.0" else get_machine_addr()

URL = f"{args.protocol}://{args.hostname}"
keyid = URL + "/actor#main-key"

if not (key := Path("privkey.pem")).exists():
	print("Creating private key. Please wait...", flush = True)
	signer = aputils.Signer.new(keyid)
	signer.export(key)

else:
	signer = aputils.Signer(key, keyid)

print(
	f"Starting server on {args.protocol}://{args.hostname} ({args.addr}:{args.port})",
	flush = True
)

server = ThreadingHTTPServer((args.addr, args.port), RequestHandler)

try:
	server.serve_forever()

except KeyboardInterrupt:
	pass



__version__ = "0.4.0"

from .errors import InvalidKeyError, SignatureFailureError
from .message import Attachment, Message, Property
from .misc import Digest, MessageDate, Signature
from .objects import HostMeta, HostMetaJson, Nodeinfo, Webfinger, WellKnownNodeinfo
from .request_classes import register_signer, register_validator
from .signer import Signer

from .enums import (
	AlgorithmType,
	NodeinfoProtocol,
	NodeinfoServiceInbound,
	NodeinfoServiceOutbound,
	NodeinfoVersion,
	ObjectType
)


__all__ = (
	"InvalidKeyError",
	"SignatureFailureError",
	"Attachment",
	"Message",
	"Property",
	"Digest",
	"MessageDate",
	"Signature",
	"HostMeta",
	"HostMetaJson",
	"NodeInfo",
	"Webfinger",
	"WellKnownNodeinfo",
	"register_signer",
	"register_validator",
	"AlgorithmType"
	"NodeinfoProtocol",
	"NodeinfoServiceInbound",
	"NodeinfoServiceOutbound",
	"NodeinfoVersion",
	"ObjectType"
)



from __future__ import annotations

import re

from blib import JsonBase
from typing import TYPE_CHECKING, Any, TypedDict

from .enums import (
	NodeinfoProtocol,
	NodeinfoServiceInbound,
	NodeinfoServiceOutbound,
	NodeinfoVersion
)

if TYPE_CHECKING:
	try:
		from typing import Self

	except ImportError:
		from typing_extensions import Self


class NewWebFingerType(TypedDict):
	subject: str
	aliases: list[str]
	links: list[dict[str, str]]


class NewNodeinfoType(TypedDict):
	version: str
	software: dict[str, str]
	protocols: list[NodeinfoProtocol]
	services: dict[str, list[Any]]
	openRegistrations: bool
	usage: dict[str, int | dict[str, int]]
	metadata: dict[str, Any] | None


class HostMeta(str):
	"An object that represents the ``/.well-known/host-meta`` endpoint"

	def __new__(cls: type[Self], data: str) -> Self:
		return str.__new__(cls, data)


	@classmethod
	def new(cls: type[Self], domain: str) -> Self:
		"""
			Generate a new host-meta xml string

			:param domain: Domain to use for the template url
		"""
		return cls(f"""<?xml version="1.0" encoding="UTF-8">
<XRD xmlns="http://docs.oasis-open.org/ns/xri/xrd-1.0">
	<Link rel="lrdd" template="https://{domain}/.well-known/webfinger?resource={{uri}}" />
</XRD>""")


class HostMetaJson(JsonBase):
	"An object that represents the ``/.well-known/host-meta.json`` endpoint"

	@classmethod
	def new(cls: type[Self], domain: str) -> Self:
		"""
			Generate a new host-meta dict

			:param domain: Domain to use for the template url
		"""
		return cls({
			"links": [
				{
					"rel": "lrdd",
					"template": f"https://{domain}/.well-known/webfinger?resource={{uri}}"
				}
			]
		})


class Nodeinfo(JsonBase):
	"An object that represents a nodeinfo endpoint"

	@classmethod
	def new(cls: type[Self],  # pylint: disable=too-many-locals,too-many-arguments
			name: str,
			version: str,
			protocols: list[NodeinfoProtocol | str] | str | None = None,
			insrv: list[NodeinfoServiceInbound | str] | None = None,
			outsrv: list[NodeinfoServiceOutbound | str] | None = None,
			metadata: dict[str, Any] | None = None,
			repo: str | None = None,
			homepage: str | None = None,
			open_regs: bool = True,
			users: int = 0,
			halfyear: int = 0,
			month: int = 0,
			posts: int = 0,
			comments: int = 0) -> Self:
		"""
			Create a new nodeinfo object. It will default to version 2.0 if ``repo`` and ``homeage``
			are not set.

			:param name: Software name to use. Can only include lowercase letters and "-", "_"
				characters.
			:param version: Version of the software.
			:param protocols: List of supported protocols
			:param insrv: Supported inbound services
			:param outsrv: Supported outbound services
			:param metadata: Extra server info
			:param repo: Url to the repository that hosts the server code
			:param homepage: Url to the homepage of the server software
			:param open_regs: Whether user registrations are open or not
			:param users: Total number of registered users
			:param halfyear: Number of active users in the past 6 months
			:param month: Number of active users in the past month
			:param posts: Total number of posts
			:param comments: Total number of comments
		"""

		if protocols is None:
			protocols = []

		if insrv is None:
			insrv = []

		if outsrv is None:
			outsrv = []

		else:
			outsrv = list(NodeinfoServiceOutbound.parse(v) for v in outsrv)

		if metadata is None:
			metadata = {}

		if not re.match("^[a-z0-9-]+$", name):
			raise ValueError("Invalid software name. Must match regex: ^[a-z0-9-]+$")

		if isinstance(protocols, str):
			protocols = [protocols]

		elif not isinstance(protocols, (list, set, tuple)):
			raise TypeError("Protocols must be a list, set, or tuple")

		data: NewNodeinfoType = {
			"version": "2.1" if repo or homepage else "2.0",
			"software": {
				"name": name,
				"version": version
			},
			"protocols": list(NodeinfoProtocol.parse(v) for v in protocols),
			"services": {
				"inbound": list(NodeinfoServiceInbound.parse(v) for v in insrv),
				"outbound": outsrv
			},
			"openRegistrations": open_regs,
			"usage": {
				"localPosts": posts,
				"localComments": comments,
				"users": {
					"total": users,
					"activeHalfyear": halfyear,
					"activeMonth": month,
				}
			},
			"metadata": metadata
		}

		if repo:
			data["software"]["repository"] = repo

		if homepage:
			data["software"]["homepage"] = homepage

		return cls(data)


	@property
	def sw_name(self) -> str:
		"Get the software name at ``Nodeinfo['software']['name']``"
		return self["software"]["name"]


	@property
	def sw_version(self) -> str:
		"Get the software version at ``Nodeinfo['software']['version']``"
		return self["software"]["version"]


class Webfinger(JsonBase):
	"An object that represents the ``/.well-known/webfinger`` endpoint"


	@classmethod
	def new(cls: type[Self],
			handle: str,
			domain: str,
			actor: str,
			profile: str | None = None,
			interaction: str | None = None) -> Self:
		"""
			Create a new webfinger object

			:param handle: Username of the account
			:param domain: Domain the account resides on
			:param actor: URL that points to the ActivityPub Actor of the account
			:param profile: URL that points to an HTML representation of the account
			:param interaction: URL used for remote interaction
		"""
		data: NewWebFingerType = {
			"subject": f"acct:{handle}@{domain}",
			"aliases": [actor],
			"links": [{
				"rel": "self",
				"type": "application/activity+json",
				"href": actor
			}]
		}

		if profile:
			data["aliases"].append(profile)
			data["links"].append({
				"rel": "http://webfinger.net/rel/profile-page",
				"type": "text/html",
				"href": profile
			})

		if interaction:
			data["links"].append({
				"rel": "http://ostatus.org/schema/1.0/subscribe",
				"template": interaction
			})

		return cls(data)


	@property
	def handle(self) -> str:
		"Username of the account"
		return self["subject"][5:].split("@", 1)[0]


	@property
	def domain(self) -> str:
		"Domain the account resides on"
		return self["subject"][5:].split("@", 1)[1]


	@property
	def profile(self) -> str:
		"URL to the HTML representation of the account"
		return self.get_link("profile")


	@property
	def actor(self) -> str:
		"URL that points to the ActivityPub Actor of the account"
		return self.get_link("actor")


	@property
	def interaction(self) -> str:
		"URL used for remote interaction"
		return self.get_link("interaction")


	def get_link(self, link_type: str) -> str:
		"""
			Get a URL for the account

			:param link_type: The type of URL to get [profile, actor, interaction]
		"""

		for link in self["links"]:
			if link_type == "profile" and link["rel"] == "http://webfinger.net/rel/profile-page":
				return link["href"]

			if link_type == "actor" and link["rel"] == "self":
				return link["href"]

			if link_type == "interaction" and link["rel"] == "http://ostatus.org/schema/1.0/subscribe":
				return link["template"]

		raise KeyError(link_type)


class WellKnownNodeinfo(JsonBase):
	"An object that represents the ``/.well-known/nodeinfo`` endpoint"

	@classmethod
	def new(cls: type[Self],
			v20: str | None = None,
			v21: str | None = None) -> Self:
		"""
			Create a new ``WellKnownNodeinfo`` object by specifying the url(s)

			:param v20: URL pointing to a nodeinfo v2.0 object
			:param v21: Url pointing to a nodeinfo v2.1 object
		"""

		if not (v20 or v21):
			raise ValueError("At least one nodeinfo version must be specified")

		data = []

		if v20:
			data.append({
				"rel": NodeinfoVersion.V20.value,
				"href": v20
			})

		if v21:
			data.append({
				"rel": NodeinfoVersion.V21.value,
				"href": v21
			})

		return cls({"links": data})


	@classmethod
	def new_template(cls: type[Self],
					domain: str,
					path: str = "/nodeinfo",
					v20: bool = True,
					v21: bool = True) -> Self:
		"""
			Create a new ``WellKnownNodeinfo`` object by specifying the domain, path, and which
			versions to include.

			:param domain: Domain to use for urls
			:param path: Base path for urls
			:param v20: If true, generate a Nodeinfo v2.0 url
			:param v21: If true, generate a Nodeinfo v2.1 url

			:raises ValueError: When both ``v20`` and ``v21`` are ``False``
		"""

		if not (v20 or v21):
			raise ValueError("At least one nodeinfo version must be specified")

		if path.endswith("/"):
			path = path[:-1]

		if not path.startswith("/"):
			path = "/" + path

		urls = {}

		if v20:
			urls["v20"] = f"https://{domain}{path}/2.0.json"

		if v21:
			urls["v21"] = f"https://{domain}{path}/2.1.json"

		return cls.new(**urls)


	@property
	def v20(self) -> str:
		"Try to get the v2.0 url. Shortcut for ``Nodeinfo.get_url('20')``"
		return self.get_url("20")


	@property
	def v21(self) -> str:
		"Try to get the v2.0 url. Shortcut for ``Nodeinfo.get_url('21')``"
		return self.get_url("21")


	def get_url(self, version: str = "20") -> str:
		"""
			Get a versioned nodeinfo url

			:param version: The nodeinfo version url to get

			:raises KeyError: If the url for the specified nodeinfo version cannot be found
		"""

		for item in self["links"]:
			if item["rel"] == NodeinfoVersion.parse("V" + version):
				return item["href"]

		raise KeyError(version)



from __future__ import annotations

from blib import StrEnum
from collections.abc import Callable
from typing import Any, Generic, TypeVar


T = TypeVar("T")
V = TypeVar("V")


class classproperty(Generic[T, V]):
	def __init__(self, func: Callable[[type[T]], V]) -> None:
		self.func: Callable[[type[T]], V] = getattr(func, "__wrapped__", func)


	def __get__(self, obj: Any, cls: type[T]) -> V:
		return self.func(cls)


class AlgorithmType(StrEnum):
	"Algorithm type"

	RSASHA256 = "rsa-sha256"
	"Old signing standard"

	HS2019 = "hs2019"
	"Current signing standard"


class NodeinfoProtocol(StrEnum):
	"Protocols for nodeinfo"

	ACTIVITYPUB = "activitypub"
	BUDDYCLOUD = "buddycloud"
	DFRN = "dfrn"
	DIASPORA = "diaspora"
	LIBERTREE = "libertree"
	OSTATUS = "ostatus"
	PUMPIO = "pumpio"
	TENT = "tent"
	XMPP = "xmpp"
	ZOT = "zot"


class NodeinfoServiceInbound(StrEnum):
	"Nodeinfo inbound services"

	ATOM = "atom1.0"
	GNUSOCIAL = "gnusocial"
	IMAP = "imap"
	PNUT = "pnut"
	POP3 = "pop3"
	PUMPIO = "pumpio"
	RSS = "rss2.0"
	TWITTER = "twitter"


class NodeinfoServiceOutbound(StrEnum):
	"Nodeinfo outbound services"

	ATOM = "atom1.0"
	BLOGGER = "blogger"
	BUDDYCLOUD = "buddycloud"
	DIASPORA = "diaspora"
	DREAMWIDTH = "dreamwidth"
	DRUPAL = "drupal"
	FACEBOOK = "facebook"
	FRIENDICA = "friendica"
	GNUSOCIAL = "gnusocial"
	GOOGLE = "google"
	INSANEJOURNAL = "insanejournal"
	LIBERTREE = "libertree"
	LINKEDIN = "linkedin"
	LIVEJOURNAL = "livejournal"
	MEDIAGOBLIN = "mediagoblin"
	MYSPACE = "myspace"
	PINTEREST = "pinterest"
	PNUT = "pnut"
	POSTEROUS = "posterous"
	PUMPIO = "pumpio"
	REDMATRIX = "redmatrix"
	RSS = "rss2.0"
	SMTP = "smtp"
	TENT = "tent"
	TUMBLR = "tumblr"
	TWITTER = "twitter"
	WORDPRESS = "wordpress"
	XMPP = "xmpp"


class NodeinfoVersion(StrEnum):
	"Namespace URLs for the different Nodeinfo versions"

	V20 = "http://nodeinfo.diaspora.software/ns/schema/2.0"
	V21 = "http://nodeinfo.diaspora.software/ns/schema/2.1"


class ObjectType(StrEnum):
	"ActivityPub object types"

	ACCEPT = "Accept"
	ADD = "Add"
	ANNOUNCE = "Announce"
	APPLICATION = "Application"
	ARRIVE = "Arrive"
	ARTICLE = "Article"
	AUDIO = "Audio"
	BLOCK = "Block"
	COLLECTION = "Collection"
	COLLECTION_PAGE = "CollectionPage"
	CREATE = "Create"
	DELETE = "Delete"
	DISLIKE = "Dislike"
	DOCUMENT = "Document"
	EMOJI = "Emoji"
	EVENT = "Event"
	FLAG = "Flag"
	FOLLOW = "Follow"
	GROUP = "Group"
	IGNORE = "Ignore"
	IMAGE = "Image"
	INVITE = "Invite"
	JOIN = "Join"
	LEAVE = "Leave"
	LIKE = "Like"
	LISTEN = "Listen"
	MOVE = "Move"
	NOTE = "Note"
	OBJECT = "Object"
	OFFER = "Offer"
	ORDERED_COLLECTION = "OrderedCollection"
	ORDERED_COLLECTION_PAGE = "OrderedCollectionPage"
	ORGANIZATION = "Organization"
	PAGE = "Page"
	PERSON = "Person"
	PLACE = "Place"
	PROFILE = "Profile"
	QUESTION = "Question"
	REJECT = "Reject"
	READ = "Read"
	RELATIONSHIP = "Relationship"
	REMOVE = "Remove"
	SERVICE = "Service"
	TENTATIVE_ACCEPT = "TentativeAccept"
	TENTATIVE_REJECT = "TentativeReject"
	TOMBSTONE = "Tombstone"
	TRAVEL = "Travel"
	UNDO = "Undo"
	UPDATE = "Update"
	VIDEO = "Video"
	VIEW = "View"


	@classproperty
	@classmethod
	def Activity(cls: type[ObjectType]) -> tuple[ObjectType, ...]:
		return (
			cls.ACCEPT,
			cls.ADD,
			cls.ANNOUNCE,
			cls.ARRIVE,
			cls.BLOCK,
			cls.CREATE,
			cls.DELETE,
			cls.DISLIKE,
			cls.FLAG,
			cls.FOLLOW,
			cls.IGNORE,
			cls.INVITE,
			cls.JOIN,
			cls.LEAVE,
			cls.LIKE,
			cls.LISTEN,
			cls.MOVE,
			cls.OFFER,
			cls.QUESTION,
			cls.REJECT,
			cls.READ,
			cls.REMOVE,
			cls.TENTATIVE_ACCEPT,
			cls.TENTATIVE_REJECT,
			cls.TRAVEL,
			cls.UNDO,
			cls.UPDATE,
			cls.VIEW
		)


	@classproperty
	@classmethod
	def Actor(cls: type[ObjectType]) -> tuple[ObjectType, ...]:
		return (
			cls.APPLICATION,
			cls.GROUP,
			cls.ORGANIZATION,
			cls.PERSON,
			cls.SERVICE
		)


	@classproperty
	@classmethod
	def Collection(cls: type[ObjectType]) -> tuple[ObjectType, ...]:
		return (
			cls.COLLECTION,
			cls.COLLECTION_PAGE,
			cls.ORDERED_COLLECTION,
			cls.ORDERED_COLLECTION_PAGE
		)


	@classproperty
	@classmethod
	def Media(cls: type[ObjectType]) -> tuple[ObjectType, ...]:
		return (
			cls.AUDIO,
			cls.EMOJI,
			cls.IMAGE,
			cls.VIDEO
		)


	@classproperty
	@classmethod
	def Object(cls: type[ObjectType]) -> tuple[ObjectType, ...]:
		return (
			cls.APPLICATION,
			cls.ARTICLE,
			cls.AUDIO,
			cls.COLLECTION,
			cls.COLLECTION_PAGE,
			cls.DOCUMENT,
			cls.EMOJI,
			cls.EVENT,
			cls.GROUP,
			cls.IMAGE,
			cls.NOTE,
			cls.OBJECT,
			cls.ORGANIZATION,
			cls.ORDERED_COLLECTION,
			cls.ORDERED_COLLECTION_PAGE,
			cls.PAGE,
			cls.PERSON,
			cls.PLACE,
			cls.PROFILE,
			cls.RELATIONSHIP,
			cls.SERVICE,
			cls.TOMBSTONE,
			cls.VIDEO
		)


class SignatureAlgorithm(StrEnum):
	"Algorithm specified in the ``algorithm`` field in the signature"

	HS2019 = "hs2019"
	RSA_SHA1 = "rsa-sha1"
	RSA_SHA256 = "rsa-sha256"
	ECDSA_SHA256 = "ecdsa-sha256"



from __future__ import annotations

import asyncio
import base64
import json
import time

from Crypto import Hash
from Crypto.PublicKey import RSA
from Crypto.Signature import PKCS1_v1_5
from blib import Date, HttpDate, JsonBase, Url
from collections.abc import Callable, Mapping, Sequence
from datetime import timedelta
from functools import wraps
from inspect import iscoroutinefunction
from pathlib import Path
from typing import TYPE_CHECKING, Any, Protocol

from .enums import AlgorithmType
from .errors import SignatureFailureError
from .misc import Digest, Signature
from .request_classes import SIGNERS, VALIDATORS

if TYPE_CHECKING:
	try:
		from typing import Self

	except ImportError:
		from typing_extensions import Self


class DataHash(Protocol):
	def update(self, data: bytes) -> None: ...
	def digest(self) -> bytes: ...
	def hexdigest(self) -> str: ...


def build_signing_string(
		headers: Mapping[str, str],
		used_headers: Sequence[str] | None = None) -> tuple[bytes, Sequence[str]]:
	"""
		Create a signing string from HTTP headers

		:param headers: Key/value pair of HTTP headers
		:param used_headers: List of headers to be used in the signing string
	"""

	if used_headers is None:
		used_headers = tuple(headers)

	parsed_headers = tuple(f"{key}: {headers[key]}" for key in used_headers)
	return "\n".join(parsed_headers).encode("ascii"), used_headers


def check_private(func: Callable) -> Callable:
	"Checks if the key is a private key before running a method"

	@wraps(func)  # noqa: ANN201
	def wrapper(key: Signer, *args: Any, **kwargs: Any) -> Any:
		if not key.is_private:
			raise TypeError(f"Cannot use method '{func.__name__}' on Signer with public key")

		return func(key, *args, **kwargs)
	return wrapper


def hash_data(data: bytes | str) -> DataHash:
	if isinstance(data, str):
		data = bytes(data, encoding = "utf-8")

	return Hash.SHA256.new(data = data)


def process_headers_for_signing(
		method: str,
		host: str | None,
		path: str,
		raw_headers: Mapping[str, str],
		body: JsonBase | Mapping[str, Any] | bytes | str | None = None,
		algo: AlgorithmType = AlgorithmType.HS2019) -> dict[str, str]:

	if body is not None:
		if isinstance(body, JsonBase):
			body = body.to_json()

		elif isinstance(body, Mapping):
			body = json.dumps(body)

		if not isinstance(body, bytes):
			body = bytes(body, encoding = "utf-8")

	headers = {"(request-target)": f"{method.lower()} {path}"}
	headers.update({key.lower(): value for key, value in raw_headers.items()})

	if "date" not in headers:
		date = HttpDate.new_utc()

	else:
		date = HttpDate.parse(headers["date"])

	if algo == AlgorithmType.HS2019:
		headers.update({
			"(created)": str(date.timestamp()),
			"(expires)": str((date + timedelta(hours=6)).timestamp())
		})

	if host:
		headers["host"] = host

	if "date" not in headers:
		headers["date"] = date.to_string()

	if body is not None:
		headers.update({
			"digest": Digest.new(body).compile(),
			"content-length": str(len(body))
		})

	return headers


class Signer:
	"Used to sign or verify HTTP headers"

	_test_mode: bool = False
	__slots__: tuple[str, ...] = ("_key", "keyid")


	def __init__(self, key: str | Path | RSA.RsaKey, keyid: str) -> None:
		"""
			Create a new signer object. The key can be an ``RsaKey`` object, ``str``, or ``Path`` to
			an exported key

			:param key: RSA key to use for signing or verifying
			:param keyid: Url to a web resource which hosts the public key
		"""

		self._key: RSA.RsaKey = None # type: ignore[assignment]
		self._parse_key(key)

		self.keyid: str = keyid
		"Url to a web resource which hosts the public key"


	def __repr__(self) -> str:
		return f"{self.__class__.__name__}(type='RSA', bits={self.bits}, keyid='{self.keyid}')"


	@classmethod
	def new(cls: type[Self], keyid: str) -> Self:
		"""
			Create a new signer with a generated ``RsaKey`` of the specified size

			:param keyid: Url to a web resource which hosts the public key
		"""

		return cls(RSA.generate(4096), keyid)


	@classmethod
	def new_from_actor(cls: type[Self], actor: dict[str, Any]) -> Self:
		"""
			Create a signer object from an ActivityPub actor dict

			:param dict actor: ActivityPub Actor object
		"""

		return cls(actor["publicKey"]["publicKeyPem"], actor["publicKey"]["id"])


	@property
	def key(self) -> RSA.RsaKey:
		"Key to use for signing or verifying"

		return self._key


	def _parse_key(self, value: RSA.RsaKey | Path | str) -> None:
		if isinstance(value, Path):
			with value.open("r") as fd:
				value = fd.read()

		if isinstance(value, str):
			if not value.startswith("-"):
				with Path(value).expanduser().resolve().open("r", encoding = "utf-8") as fd:
					value = fd.read()

			try:
				value = RSA.import_key(value)

			except ValueError:
				raise TypeError("Invalid RSA key") from None

		if not isinstance(value, RSA.RsaKey):
			raise TypeError(
				"Key must be an RsaKey, Path, or a string representation of a key"
			)

		self._key = value


	@property
	def bits(self) -> int:
		"Size of the RSA key in bits"

		return self.key.size_in_bits()


	@property
	def is_private(self) -> bool:
		"Return ``True`` if the key is private"

		return self.key.has_private()


	@property
	@check_private
	def pubkey(self) -> str:
		"Export the public key to a str"

		key_data = self.key.public_key().export_key(format="PEM")
		return key_data.decode("utf-8") if isinstance(key_data, bytes) else key_data


	def export(self, path: Path | str | None = None) -> str:
		"""
			Export the key to a str

			:param path: Path to dump the key in text form to if specified
		"""

		key_data = self.key.export_key(format = "PEM")
		key = key_data.decode("utf-8") if isinstance(key_data, bytes) else key_data

		if path:
			path = Path(path)

			with path.open("w", encoding = "utf-8") as fd:
				fd.write(key)

		return key


	@check_private
	def sign_headers(self,
					method: str,
					url: Url | str,
					body: dict[str, Any] | bytes | str | None = None,
					headers: dict[str, str] | None = None,
					algorithm: AlgorithmType = AlgorithmType.HS2019) -> dict[str, Any]:
		"""
			Generate a signature and return the headers with a "Signature" key

			Note: HS2019 is the only supported algorithm, so only use others when you absolutely
			have to

			:param method: HTTP method of the request
			:param url: URL of the request
			:param body: ActivityPub message for a POST request
			:param headers: Request headers
			:param algorithm: Type of algorithm to use for hashing the headers. HS2019 is the only
				non-deprecated algorithm.
		"""

		if not isinstance(url, Url):
			url = Url.parse(url)

		headers = process_headers_for_signing(
			method, url.hostname, url.path, headers or {}, body, algorithm
		)

		hash_bytes, used_headers = build_signing_string(headers, None)
		signed_data = PKCS1_v1_5.new(self.key).sign(hash_data(hash_bytes))
		signature = Signature(
			base64.b64encode(signed_data).decode("utf-8"),
			self.keyid,
			algorithm,
			used_headers
		)

		if algorithm == AlgorithmType.HS2019:
			signature.created = Date.parse(int(headers["(created)"]))
			signature.expires = Date.parse(int(headers["(expires)"]))

		for key in {"(request-target)", "(created)", "(expires)", "host"}:
			headers.pop(key, None)

		headers["signature"] = signature.compile()
		return headers


	@check_private
	def sign_request(self, request: Any, algorithm: AlgorithmType = AlgorithmType.HS2019) -> Any:
		"""
			Convenience function to sign a request.

			Supported frameworks:

			* `Urllib <https://docs.python.org/3/library/urllib.request.html>`_
			* `Requests <https://pypi.org/project/requests>`_

			:param request: Request object to sign
			:param algorithm: Type of algorithm to use for signing and hashing the headers. HS2019
				is the only non-deprecated algorithm.

			:raises TypeError: If the Request class is not supported
		"""

		for rtype, func in SIGNERS.items():
			if not isinstance(request, rtype):
				continue

			if iscoroutinefunction(func):
				raise TypeError(f"Signer function cannot be a coroutine: {func.__name__}")

			return func(self, request, algorithm)

		raise TypeError(f"Request from module not supported: {type(request).__module__}")


	def validate_signature(self,
						method: str,
						path: str,
						headers: Mapping[str, Any],
						body: bytes | str | None = None) -> bool:
		"""
			Check to make sure the Signature and Digest headers match

			:param method: Request method
			:param path: Request path
			:param headers: Request headers
			:param body: Request body if it exists

			:raises aputils.SignatureFailureError: When any step of the validation process fails
		"""

		if not (signature := Signature.new_from_headers(headers)):
			raise SignatureFailureError("Missing 'signature' header")

		headers = {key.lower(): value for key, value in headers.items()}

		for key in signature.headers:
			if not key.startswith("(") and key not in headers:
				raise SignatureFailureError(f"Header key does not exist: {key}")

		headers["(request-target)"] = f"{method.lower()} {path}"

		if signature.created is not None:
			if not self._test_mode and signature.created > Date.new_utc():
				raise SignatureFailureError("Signature creation date is in the future")

			headers["(created)"] = str(signature.created.timestamp())

		if signature.expires is not None:
			if not self._test_mode and signature.expires < Date.new_utc():
				raise SignatureFailureError("Signature has expired")

			headers["(expires)"] = str(signature.expires.timestamp())

		if (digest := Digest.parse(headers.get("digest"))) is not None:
			if body is None:
				raise SignatureFailureError("A digest was added with an empty body")

			if not digest.validate(body):
				raise SignatureFailureError("Body digest does not match")

		sig_hash, _ = build_signing_string(headers, signature.headers)

		return PKCS1_v1_5.new(self.key).verify(
			hash_data(sig_hash),
			base64.b64decode(signature.signature)
		)


	def validate_request(self, request: Any) -> bool:
		"""
			Validate the signature of a server request object.

			Supported frameworks:

			* `AioHTTP <https://pypi.org/project/aiohttp>`_
			* `Flask <https://pypi.org/project/Flask>`_
			* `Sanic <https://pypi.org/project/sanic>`_

			:param request: Server request to validate
		"""

		for rtype, func in VALIDATORS.items():
			if not isinstance(request, rtype):
				continue

			if not iscoroutinefunction(func):
				return func(self, request)

			task = asyncio.create_task(func(self, request))
			elapsed = 0.0

			while not task.done():
				time.sleep(0.1)
				elapsed += 0.1

				if elapsed >= 5.0:
					task.cancel()
					raise TimeoutError(f"Validator function took too long: {func.__name__}")

			return task.result()

		raise TypeError(f"Unsupported request type: {type(request).__name__}")


	async def validate_request_async(self, request: Any) -> bool:
		"""
			Validate the signature of a server request object. Uses async when possible.

			:param request: Server request to validate
		"""

		for rtype, func in VALIDATORS.items():
			if not isinstance(request, rtype):
				continue

			if not iscoroutinefunction(func):
				return func(self, request)

			return await func(self, request)

		raise TypeError(f"Unsupported request type: {type(request).__name__}")



class InvalidKeyError(KeyError):
	"Raise when an invalid key is attempting to be added via the data parameter in Message.new"

	@property
	def key(self) -> str:
		"Name of the invalid key"

		return self.args[0]


	def __str__(self) -> str:
		return f"Invalid message key '{self.args[0]}'"


class SignatureFailureError(Exception):
	"Raise when a signature could not be verified for any reason"



from __future__ import annotations

import base64
import json

from Crypto.Hash import SHA256, SHA512
from blib import Date
from collections.abc import Mapping, Sequence
from datetime import datetime
from typing import Any, TYPE_CHECKING

from .enums import AlgorithmType

if TYPE_CHECKING:
	try:
		from typing import Self

	except ImportError:
		from typing_extensions import Self


HASHES = {
	"sha256": SHA256,
	"sha512": SHA512
}


class Digest:
	"Represents a body digest"

	__slots__ = {"digest", "algorithm"}

	def __init__(self, digest: str, algorithm: str) -> None:
		"""
			Create a new digest object from an already existing digest

			:param digest: Base64 encoded hash
			:param algorithm: Algorithm used for hash in the format of ``{type}-{bytes}``
		"""

		self.digest = digest
		self.algorithm = algorithm


	def __repr__(self) -> str:
		return f"Digest({self.digest}, {self.algorithm})"


	def __str__(self) -> str:
		return self.compile()


	def __eq__(self, value: object) -> bool:
		if isinstance(value, Digest):
			return self.algorithm == value.algorithm and self.digest == value.digest

		if isinstance(value, (dict, str, bytes)):
			return self.validate(value)

		raise TypeError(f"Cannot compare '{type(value).__name__}' to 'Digest'")


	@classmethod
	def new(cls: type[Self],
			body: dict[str, Any] | str | bytes,
			size: int = 256) -> Self:
		"""
			Generate a new body digest

			:param body: Message body to hash
			:param size: SHA hash size
		"""

		if isinstance(body, dict):
			body = json.dumps(body).encode("utf-8")

		elif isinstance(body, str):
			body = body.encode("utf-8")

		if size >= 1024:
			size = int(size / 8)

		raw_hash = HASHES[f"sha{size}"].new(data=body)
		digest = base64.b64encode(raw_hash.digest()).decode("utf-8")

		return cls(digest, f"SHA-{size}")


	@classmethod
	def parse(cls: type[Self], digest: str | None) -> Self | None:
		"""
			Create a new digest from a digest header

			:param digest: Digest header
		"""

		if not digest:
			return None

		alg, digest = digest.split("=", 1)
		return cls(digest, alg)


	@property
	def hashalg(self) -> str:
		"Hash function used when creating the signature as a string"

		return self.algorithm.replace("-", "").lower()


	def compile(self) -> str:
		"Turn the digest object into a ``str`` for the Digest header"

		return "=".join([self.algorithm, self.digest])


	def validate(self, body: dict[str, Any] | str | bytes, hash_size: int = 256) -> bool:
		"""
			Check if the body digest matches the body

			:param body: Message body to verify
			:param hash_size: Size of the hashing algorithm
		"""

		return self == Digest.new(body, hash_size)


class MessageDate(Date):
	"""
		Datetime object with convenience methods for parsing and creating ActivityPub message date
		strings
	"""

	FORMAT: str = "%Y-%m-%dT%H:%M:%SZ"
	ALT_FORMATS = ["%Y-%m-%dT%H:%M:%S.%fZ"]


class Signature:
	"""
		Represents a signature header value
	"""

	__slots__ = {"keyid", "algorithm", "headers", "signature", "created", "expires"}

	def __init__(self,
				signature: str,
				keyid: str,
				algorithm: AlgorithmType | str,
				headers: Sequence[str] | str,
				created: Date | datetime | str | int | float | None = None,
				expires: Date | datetime | str | int | float | None = None) -> None:
		"""
			Create a new signature object. This should not be initiated directly.

			:param signature: Generated signature hash
			:param keyid: URL of the public key
			:param algorithm: Hashing and signing algorithms used to create the signature
			:param headers: Header keys used to create the signature
			:param created: Unix timestamp representing the signature creation date
			:param expires: Unix timestamp representing the date the signature expires
		"""

		self.signature: str = signature
		"Generated signature hash"

		self.keyid: str = keyid
		"URL of the public key"

		self.algorithm: AlgorithmType = AlgorithmType.parse(algorithm)
		"Hashing and signing algorithms used to create the signature"

		self.headers: Sequence[str] = headers.split() if isinstance(headers, str) else headers
		"Header keys used to create the signature"

		self.created: Date | None = None
		"Signature creation date"

		self.expires: Date | None = None
		"Signature expiration date"

		self.set_created(created)
		self.set_expires(expires)


	def __repr__(self) -> str:
		data = {
			"keyid": repr(self.keyid),
			"algorithm": self.algorithm,
			"headers": self.headers,
			"created": self.created.timestamp() if self.created is not None else None,
			"expires": self.expires.timestamp() if self.expires is not None else None
		}

		str_data = ", ".join(f"{key}={value}" for key, value in data.items())
		return f"Signature({str_data})"


	@classmethod
	def new_from_headers(cls: type[Self], headers: Mapping[str, str]) -> Self:
		"""
			Parse the signature from a header dict

			:param dict[str,str] headers: Header key/value pairs
			:raises KeyError: When the signature header(s) cannot be found
			:raises NotImplementedError: When a newer unsupported signature standard is provided
		"""

		headers = {key.lower(): value for key, value in headers.items()}

		if "signature-input" in headers:
			raise NotImplementedError("Newer signature spec not supported yet")

		signature = headers["signature"]
		data: dict[str, str] = {}

		for chunk in signature.strip().split(","):
			key, value = chunk.split("=", 1)
			data[key.lower()] = value.strip("\"")

		sig = cls(
			signature = data["signature"],
			keyid = data["keyid"],
			algorithm = data["algorithm"],
			headers = data["headers"]
		)

		sig.set_created(data.get("created"))
		sig.set_expires(data.get("expires"))

		return sig


	@classmethod
	def parse(cls: type[Self], data: str) -> Self:
		"""
			Parse a Signature in string format

			:param str data: Signature string
		"""

		return cls.new_from_headers({"signature": data})


	def _set_date(self, key: str, value:  Date | datetime | str | int | float | None) -> None:
		if value is None:
			setattr(self, key, None)
			return

		if isinstance(value, str):
			try:
				value = int(value)

			except ValueError:
				pass

		setattr(self, key, Date.parse(value))


	def set_created(self, value: Date | datetime | str | int | float | None) -> None:
		"""
			Set the ``created`` property.

			:param value: Date value to be parsed
		"""

		self._set_date("created", value)


	def set_expires(self, value: Date | datetime | str | int | float | None) -> None:
		"""
			Set the ``expires`` property.

			:param value: Date value to be parsed
		"""

		self._set_date("expires", value)


	def compile(self) -> str:
		"Generate a string for a Signature header"

		data = {
			"keyId": self.keyid,
			"algorithm": self.algorithm.value,
			"headers": " ".join(self.headers),
			"created": self.created.timestamp() if self.created is not None else None,
			"expires": self.expires.timestamp() if self.expires is not None else None,
			"signature": self.signature
		}

		return ",".join([f"{k}=\"{v}\"" for k, v in data.items() if v is not None])



from __future__ import annotations

import json
import re
import typing

from blib import Date, HttpDate, JsonBase
from collections.abc import Callable
from datetime import datetime
from functools import cached_property
from mimetypes import guess_type
from typing import TYPE_CHECKING, Any, TypeVar
from urllib.parse import urlparse

from .enums import ObjectType
from .errors import InvalidKeyError
from .misc import MessageDate
from .signer import Signer

if TYPE_CHECKING:
	try:
		from typing import Self

	except ImportError:
		from typing_extensions import Self


T = TypeVar("T")
TYPES: dict[str, type | tuple[type, ...]] = {
	"str": str,
	"int": int,
	"float": float,
	"bool": bool,
	"date": datetime,
	"dict": dict,
	"object": (dict, str),
	"list": (list, set, tuple)
}

CONVERTERS: dict[type, tuple[Callable, Callable]] = {
	datetime: (MessageDate.parse, lambda v: v.strftime(MessageDate.FORMAT)),
	Date: (Date.parse, Date.to_string),
	HttpDate: (HttpDate.parse, HttpDate.to_string),
	MessageDate: (MessageDate.parse, MessageDate.to_string),
	dict: (JsonBase.parse, lambda v: JsonBase(v).to_json()),
	JsonBase: (JsonBase.parse, JsonBase.to_json),
	list: (json.loads, json.dumps)
}


class Property(typing.Generic[T]):
	"Represents a key in a dict"


	def __init__(self, type_name: str = "str") -> None:
		"""
			Create a new dict property

			:param type_name: Name of the value type to be used for (de)serialization
		"""

		self.key: str = ""
		self.type_name: str = type_name


	def __set_name__(self, obj: Any, key: str) -> None:
		self.key = Property.name_to_camel_case(key)


	def __get__(self, obj: Message | None, objtype: Any | None = None) -> Any:
		if obj is None:
			return self

		try:
			return obj[self.key]

		except KeyError:
			objname = object.__class__.__name__
			raise AttributeError(f"'{objname}' has no attribute '{self.key}'") from None


	def __set__(self, obj: Message, value: Any) -> None:
		if type(value) is dict:
			value = Message(value)

		obj[self.key] = value


	def __delete__(self, obj: Message) -> None:
		del obj[self.key]


	@staticmethod
	def name_to_snake_case(key: str) -> str:
		new_key = "_".join(re.split(r"(?<=.)(?=[A-Z])", key))
		return new_key.lower()


	@staticmethod
	def name_to_camel_case(key: str) -> str:
		if "_" not in key:
			return key

		parts = key.split("_")

		if len(parts) > 1:
			return "".join([parts[0], *(part.title() for part in parts[1:])])

		return parts[0]


	@classmethod
	def add_type(self,
				name: str,
				cls: type,
				deserializer: Callable,
				serializer: Callable) -> None:
		"""
			Add or update a value type

			:param name: Name of the value type
			:param cls: Class of the deserialized value
			:param serializer: Method that will turn the value into a JSON-friendly type
			:param deserializer: Method that will convert the value to the specified class
		"""

		TYPES[name] = cls
		CONVERTERS[cls] = (deserializer, serializer)


class Message(JsonBase[Any]):
	":class:`dict` object that represents an ActivityPub Object"

	__slots__: tuple[str, ...] = tuple([])


	id: Property[str] = Property("str")
	"``id`` key"

	type: Property[str] = Property("str")
	"``type`` key"

	accuracy: Property[float] = Property("float")
	"``accuracy`` key"

	actor: Property[str] | Message = Property("object")
	"``actor`` key"

	altitude: Property[float] = Property("float")
	"``altitude`` key"

	any_of: Property[dict[str, Any]] = Property("dict")
	"``anyOf`` key"

	attachment: Property[list[Attachment]] = Property("list")
	"``attachment`` key"

	attributed_to: Property[str | Message] = Property("object")
	"``attributedTo`` key"

	audience: Property[str | Message] = Property("object")
	"``audience`` key"

	bcc: Property[list[str]] = Property("list")
	"``bcc`` key"

	bto: Property[list[str]] = Property("list")
	"``bto`` key"

	cc: Property[list[str]] = Property("list")
	"``cc`` key"

	closed: Property[MessageDate] = Property("date")
	"``closed`` key"

	content: Property[str] = Property("str")
	"``content`` key"

	content_map: Property[dict] = Property("dict")
	"``contentMap`` key"

	context: Property[str] = Property("str")
	"``context`` key"

	created: Property[MessageDate] = Property("date")
	"``created`` key"

	current: Property[str] | Message = Property("object")
	"``current`` key"

	deleted: Property[MessageDate] = Property("date")
	"``deleted`` key"

	describes: Property[str] | Message = Property("object")
	"``describes`` key"

	devices: Property[str] = Property("str")
	"``devices`` key"

	discoverable: Property[bool] = Property("bool")
	"``discoverable`` key"

	duration: Property[int] = Property("str")
	"``duration`` key"

	end_time: Property[MessageDate] = Property("date")
	"``endTime`` key"

	endpoints: Property[str | Message] = Property("object")
	"``endpoints`` key"

	featured: Property[str] = Property("str")
	"``featured`` key"

	featured_tags: Property[str] = Property("str")
	"``featuredTags`` key"

	first: Property[str] = Property("str")
	"``first`` key"

	followers: Property[str] = Property("str")
	"``followers`` key"

	following: Property[str] = Property("following")
	"``following`` key"

	former_type: Property[str] = Property("str")
	"``formerType`` key"

	generator: Property[str | Message] = Property("object")
	"``generator`` key"

	height: Property[int] = Property("int")
	"``height`` key"

	href: Property[str] = Property("str")
	"``href`` key"

	href_lang: Property[str] = Property("str")
	"``hreflang`` key"

	icon: Property[str | Message] = Property("object")
	"``icon`` key"

	image: Property[str | Message] = Property("object")
	"``image`` key"

	in_reply_to: Property[str | Message] = Property("object")
	"``inReplyTo`` key"

	inbox: Property[str] = Property("str")
	"``inbox`` key"

	instrument: Property[str | Message] = Property("object")
	"``instrument`` key"

	obj_items: Property[dict[str, Any]] = Property("str")
	"``items`` key"

	last: Property[str] = Property("str")
	"``last`` key"

	latitude: Property[float] = Property("float")
	"``latitude`` key"

	link: Property[str | Message] = Property("object")
	"``link`` key"

	location: Property[str | Message] = Property("object")
	"``location`` key"

	longitude: Property[float] = Property("float")
	"``longitude`` key"

	manually_approves_followers: Property[bool] = Property("bool")
	"``manuallyApprovesFollowers`` key"

	media_type: Property[str] = Property("str")
	"``mediaType`` key"

	name: Property[str] = Property("str")
	"``name`` key"

	next: Property[str] = Property("str")
	"``next`` key"

	object: Property[str | Message] = Property("object")
	"``object`` key"

	one_of: Property[dict[str, Any]] = Property("str")
	"``oneOf`` key"

	origin: Property[str | Message] = Property("object")
	"``origin`` key"

	outbox: Property[str] = Property("str")
	"``outbox`` key"

	owner: Property[str] = Property("str")
	"``owner`` key"

	part_of: Property[str] = Property("str")
	"``partOf`` key"

	preferred_username: Property[str] = Property("str")
	"``preferredUsername`` key"

	preview: Property[str | Message] = Property("object")
	"``preview`` key"

	previous: Property[str] = Property("str")
	"``previous`` key"

	public_key: Property[str | Message] = Property("object")
	"``publicKey`` key"

	public_key_pem: Property[str] = Property("str")
	"``publicKeyPem`` key"

	published: Property[MessageDate] = Property("date")
	"``published`` key"

	radius: Property[int] = Property("int")
	"``radius`` key"

	rel: Property[str] = Property("str")
	"``rel`` key"

	relationship: Property[str | Message] = Property("object")
	"``relationship`` key"

	replies: Property[str | Message] = Property("object")
	"``replies`` key"

	result: Property[str | Message] = Property("object")
	"``result`` key"

	start_index: Property[int] = Property("int")
	"``startIndex`` key"

	start_time: Property[MessageDate] = Property("date")
	"``startTime`` key"

	summary: Property[str] = Property("str")
	"``summary`` key"

	tag: Property[dict[str, Any]] = Property("object")
	"``tag`` key"

	target: Property[str | Message] = Property("object")
	"``target`` key"

	to: Property[list[str]] = Property("list")
	"``to`` key"

	total_items: Property[int] = Property("int")
	"``totalItems`` key"

	units: Property[str] = Property("str")
	"``units`` key"

	updated: Property[MessageDate] = Property("date")
	"``updated`` key"

	url: Property[str] = Property("str")
	"``url`` key"

	value: Property[str] = Property("str")
	"``value`` key"

	width: Property[int] = Property("int")
	"``width`` key"


	def __getitem__(self, key: str) -> Any:
		return dict.__getitem__(self, Property.name_to_camel_case(key))


	def __setitem__(self, key: str, value: Any) -> None:  # noqa: ANN001
		dict_key = Property.name_to_camel_case(key)
		key = Property.name_to_snake_case(key)

		if key == "attachment":
			value = [Attachment(item) for item in value] if value else []

		else:
			try:
				prop = getattr(Message, key)
				vtype = TYPES[prop.type_name]

				if isinstance(value, datetime) and not isinstance(value, MessageDate):
					value = MessageDate.parse(value)

				elif vtype == TYPES["object"] and isinstance(value, str):
					pass

				else:
					vtype = vtype[0] if isinstance(vtype, tuple) else vtype

					if not isinstance(value, vtype) and value is not None:
						value = CONVERTERS[vtype][0](value)

			except (KeyError, AttributeError):
				pass

		dict.__setitem__(self, dict_key, value)


	def __delitem__(self, key: str) -> None:
		dict.__delitem__(self, Property.name_to_camel_case(key))


	@classmethod
	def new(cls,
			obj_type: ObjectType | str,
			data: dict[str, Any],
			context: list[Any] | None = None) -> Self:
		"""
			Create a new ActivityPub object

			:param obj_type: Type of object to create
			:param data: Key/value pairs to include in the message
			:param context: List of JSON-LD context items
			:raises InvalidKeyError: When an invalid key has been specified
		"""

		if not context:
			context = []

		elif not isinstance(context, (list, tuple, set)):
			raise TypeError("Object context must be a list, tuple, or set")

		if "https://www.w3.org/ns/activitystreams" not in context:
			context.insert(0, "https://www.w3.org/ns/activitystreams")

		message = cls({"@context": context})
		message.type = ObjectType.parse(obj_type)

		for key, value in (data or {}).items():
			if key in {"@context", "type"}:
				raise InvalidKeyError(key)

			message[key] = value

		return message


	@classmethod
	def new_image(cls, url: str, mimetype: str | None = None) -> Self:
		"""
			Create a new ``Image`` object. If a mimetype is not provided, it will be guessed based
			on the filename in the url.

			:param url: Location of the image
			:param mimetype: Mimetype to use
			:raises ValueError: When a mimetype is not provided and cannot be determined
		"""

		if not mimetype:
			if not (mimetype := guess_type(urlparse(url).path)[0]):
				raise ValueError(f"Cannot determine mimetype of url: {url}")

		data = cls()
		data.type = ObjectType.IMAGE
		data.media_type = mimetype
		data.url = url
		return data


	# general properties
	@property
	def actor_id(self) -> str:
		"Get the url of the actor associated with the message"

		if "actor" not in self:
			raise AttributeError("Message does not have an 'actor' key")

		if isinstance(self["actor"], str):
			return self["actor"]

		try:
			return self["actor"]["id"]

		except KeyError:
			raise AttributeError("Message does not have an 'actor' key") from None


	@property
	def domain(self) -> str:
		"Get the domain of the object origin"
		return urlparse(self["id"]).hostname


	@property
	def object_id(self) -> str:
		"Get the ``id`` field of a linked object if it exists"

		if "object" not in self:
			raise AttributeError("Message does not have an 'object' key")

		if isinstance(self["object"], str):
			return self["object"]

		try:
			return self["object"]["id"]

		except KeyError:
			raise AttributeError("Message does not have an 'object' key") from None


	@property
	def object_domain(self) -> str:
		"Get the domain of the linked object if it exists"

		if (domain := urlparse(self.object_id).hostname) is None:
			raise ValueError("Could not get domain")

		return domain


	def add_field(self, key: str, value: str) -> None:
		"""
			Add a profile field

			:param key: Name of the field
			:param value: Value to associate with the name
		"""

		if "attachment" not in self:
			self.attachment = []

		self.attachment.append(Attachment.new_field(key, value))


	def add_image(self, key: str, url: str) -> None:
		"""
			Add an ``Image`` object at the specified key

			:param key: Key name to attach the ``Image`` object to
			:param url: Location of the image
		"""
		self[key] = type(self).new_image(url)


	def del_field(self, key: str) -> None:
		"""
			Delete a profile field

			:param key: Name of the field
		"""
		self.attachment.remove(self.get_field(key))


	def get_field(self, key: str) -> dict[str, str]:
		"""
			Get a field dict with the specified name

			:param key: Name of the field
		"""
		for field in self.get("attachment", []):
			if field["type"] == "PropertyValue" and field["name"] == key:
				return field

		raise KeyError(key)


	def get_fields(self) -> dict[str, str]:
		"Get all profile fields as key/value pairs"
		data = {}

		for field in self.get("attachment", []):
			if field["type"] == "PropertyValue":
				data[field["name"]] = field["value"]

		return data


	# Actor stuff
	@classmethod
	def new_actor(cls,
				actor_type: ObjectType,
				handle: str,
				actor: str,
				pubkey: str,
				inbox: str | None = None,
				outbox: str | None = None,
				shared_inbox: str | None = None,
				fields: dict[str, str] | None = None,
				avatar: str | None = None,
				header: str | None = None,
				**kwargs: Any) -> Self:
		"""
			Create a new actor object

			:param actor_type: Type of actor to create
			:param handle: Username of the actor
			:param actor: URL to where this object will be hosted
			:param pubkey: PEM of the public key associated with the actor
			:param inbox: URL to the actor's inbox
			:param outbox: URL to the actor's outbox,
			:param shared_inbox: URL to the inbox shared amongst all users of the instance
			:param fields: key/value string pairs to display on the actor's profile page
			:param avatar: URL to an image to be used as the actor's profile picture
			:param header: URL to an image to be used as the actor's header image
			:param kwargs: Extra object values to set
			:raises ValueError: When a non-actor type is provided
		"""

		if (actor_type := ObjectType.parse(actor_type)) not in ObjectType.Actor:
			raise ValueError(f"Invalid Actor type: {actor_type.value}")

		parsed_url = urlparse(actor)
		proto = parsed_url.scheme
		domain = parsed_url.netloc

		data = cls.new(
			actor_type,
			data = {
				"id": actor,
				"preferred_username": handle,
				"inbox": inbox or f"{actor}/inbox",
				"outbox": outbox or f"{actor}/outbox",
				"attachment": [],
				**kwargs,
				"endpoints": {
					"sharedInbox": shared_inbox or f"{proto}://{domain}/inbox"
				},
				"public_key": {
					"id": f"{actor}#main-key",
					"owner": actor,
					"publicKeyPem": pubkey
				}
			},
			context = [
				{
					"toot": "http://joinmastodon.org/ns#",
					"publicKeyBase64": "toot:publicKeyBase64",
				}
			]
		)

		if fields:
			data["@context"][1].update({
				"schema": "http://schema.org#",
				"PropertyValue": "schema:PropertyValue",
				"value": "schema:value"
			})

			for key, value in fields.items():
				data.add_field(key, value)

		else:
			del data.attachment

		if avatar:
			data.add_image("icon", avatar)

		if header:
			data.add_image("image", header)

		return data


	@property
	def username(self) -> str:
		"Get the actor's username (alias of :attr:`Message.preferred_username`)"
		return self["preferredUsername"]


	@property
	def keyid(self) -> str:
		"Get the ID of the actor's public key"
		return self["publicKey"]["id"]


	@property
	def pubkey(self) -> str:
		"Get the actor's PEM encoded public key"
		return self["publicKey"]["publicKeyPem"]


	@property
	def shared_inbox(self) -> str:
		"Get the instance's shared inbox"
		return self["endpoints"]["sharedInbox"]


	@cached_property
	def signer(self) -> Signer:
		"Create :class:`Signer` object for the actor"
		return Signer.new_from_actor(self)


	@property
	def handle(self) -> str:
		"Get the full username that includes the domain"
		return f"{self.username}@{self.domain}"


	def handle_value_dump(self, value: Any) -> Any:
		try:
			return CONVERTERS[type(value)][1](value)

		except KeyError:
			pass

		return JsonBase.handle_value_dump(self, value)


class Attachment(JsonBase):
	"Represents an object in the ``attachment`` list"

	@classmethod
	def new_field(cls: type[Attachment], key: str, value: str) -> Attachment:
		"""
			Create a new attachment meant for storing a key/value pair on an actor

			:param key: Name of the field
			:param value: Value to store with the field
		"""
		return cls({
			"type": "PropertyValue",
			"name": key,
			"value": value
		})



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any, Protocol
from urllib.request import Request as UrllibRequest

from .enums import AlgorithmType

if TYPE_CHECKING:
	try:
		from .signer import Signer

	except ImportError:
		class Signer(Protocol): # type: ignore
			def sign_headers(self,
							method: str,
							url: str,
							data: dict[str, Any] | bytes | str | None = None,
							headers: dict[str, str] | None = None,
							algorithm: AlgorithmType = AlgorithmType.HS2019,
							sign_all: bool = False) -> dict[str, Any]: ...

			def validate_signature(self,
									method: str,
									path: str,
									headers: dict[str, Any],
									body: bytes | str | None = None) -> bool: ...

	SignerFunc = Callable[[Signer, Any, AlgorithmType], Any]
	ValidatorFunc = Callable[[Signer, Any], bool]


try:
	from requests import Request as RequestsRequest
	from requests import PreparedRequest as RequestsPreparedRequest

except ImportError:
	RequestsRequest = None # type: ignore[misc,assignment]
	RequestsPreparedRequest = None # type: ignore[misc,assignment]


try:
	from aiohttp.web import BaseRequest as AiohttpRequest

except ImportError:
	AiohttpRequest = None # type: ignore[misc,assignment]

try:
	from flask import Request as FlaskRequest

except ImportError:
	FlaskRequest = None # type: ignore[misc,assignment]

try:
	from sanic.request import Request as SanicRequest

except ImportError:
	SanicRequest = None # type: ignore[misc,assignment]


SIGNERS: dict[type[Any], SignerFunc] = {}
VALIDATORS: dict[type[Any], ValidatorFunc] = {}


def register_signer(request_type: type[Any]) -> Callable:
	"""
		Register a function to handle signing of a request type.

		:param request_type: Class of the request
	"""

	def wrapper(func: SignerFunc) -> SignerFunc:
		SIGNERS[request_type] = func
		return func

	return wrapper


def register_validator(request_type: type[Any]) -> Callable:
	"""
		Register a function to handle verification of a request type.

		:param request_type: Class of the request
	"""

	def wrapper(func: ValidatorFunc) -> ValidatorFunc:
		VALIDATORS[request_type] = func
		return func

	return wrapper


### Signers

@register_signer(UrllibRequest)
def sign_urllib(
			signer: Signer,
			request: UrllibRequest,
			algorithm: AlgorithmType) -> UrllibRequest:

	request_headers = dict(request.header_items())
	headers = signer.sign_headers(
		request.get_method().upper(),
		request.full_url, request.data,
		request_headers,
		algorithm = algorithm
	)

	request.headers = {key.title(): value for key, value in headers.items()}
	return request


if RequestsRequest is not None:
	@register_signer(RequestsRequest)
	def sign_requests(
				signer: Signer,
				request: RequestsRequest,
				algorithm: AlgorithmType) -> RequestsPreparedRequest:

		return sign_requests_prepared(signer, request.prepare(), algorithm)


if RequestsPreparedRequest is not None:
	@register_signer(RequestsPreparedRequest)
	def sign_requests_prepared(
				signer: Signer,
				request: RequestsPreparedRequest,
				algorithm: AlgorithmType) -> RequestsPreparedRequest:

		headers = signer.sign_headers(
			request.method,
			request.url,
			request.body,
			request.headers,
			algorithm = algorithm
		)

		request.headers = headers
		return request


### Validators

if AiohttpRequest is not None:
	@register_validator(AiohttpRequest)
	async def validate_aiohttp(signer: Signer, request: AiohttpRequest) -> bool:
		return signer.validate_signature(
			request.method,
			request.path,
			{key: value for key, value in request.headers.items()},
			await request.read()
		)


if FlaskRequest is not None:
	@register_validator(FlaskRequest)
	def validate_flask(signer: Signer, request: FlaskRequest) -> bool:
		return signer.validate_signature(
			request.method,
			request.path,
			{key: value for key, value in request.headers.items()},
			request.get_data()
		)


if SanicRequest is not None:
	@register_validator(SanicRequest)
	def validate_sanic(signer: Signer, request: SanicRequest) -> bool:
		return signer.validate_signature(
			request.method,
			request.path,
			{key: value for key, value in request.headers.items()},
			request.body
		)
